"""
Multivariate Polynomials
"""

import weakref

from ring import Ring

import multi_polynomial
import polylist_fcns
import sage.structure.gens as gens


class MPolynomialRing(object):
    def __new__(cls, *args, **kwds):
        if cls != MPolynomialRing:
            return object.__new__(cls, *args, **kwds)
        if len(args) == 0:
            raise TypeError, "MPolynomialRing requires at least 1 argument."
        return MPolynomialRing_polylist(*args, **kwds)

    def __init__(self, *args, **kwds):
        raise NotImplementedError

class MPolynomialRing_base(Ring, gens.Gens):
    """
    Multivariate polynomial ring.

    EXAMPLES:

    """
    def __init__(self, base_ring, n=1, variable="x"):
        Ring.__init__(self)
        if not isinstance(base_ring, Ring):
            raise TypeError, "Base ring (=%s) must be a Ring."%base_ring
        self.__base_ring = base_ring
        n = int(n)
        if n <= 0:
            raise ValueError, "Multivariate Polynomial rings must " + \
                  "have more than 0 variables."
        self.__ngens = n
        if not isinstance(variable, str):
            raise TypeError, "Variable (=%s) must be a string."%variable
        self.__variable = variable
        self.__var_names = ["%s%s"%(variable,i) for i in range(n)]
        self.__repr = "Polynomial ring in %s over %s"%(
            ", ".join(self.__var_names), base_ring)

    def __repr__(self):
        return self.__repr

    def base_ring(self):
        return self.__base_ring

    def ngens(self):
        return self.__ngens

    def variable(self):
        return self.__variable

    def _monomial_order_function(self):
        raise NotImplementedError


class MPolynomialRing_polylist(MPolynomialRing_base):

    def __init__(self, base_ring, n=1, variable="x"):
        self.__monomial_order_function = polylist_fcns.lex_monord
        MPolynomialRing_base.__init__(self, base_ring, n, variable)
        # Construct the generators
        v = [base_ring(0) for _ in xrange(n)]
        one = base_ring(1); zero = base_ring(0)
        self.__gens = []
        for i in xrange(n):
            v[i] = one
            self.__gens.append(multi_polynomial.MPolynomial_polylist(self, [[one, list(v)]]))
            v[i] = zero
        self.__zero_list = v
            
    def gen(self, n=0):
        if n < 0 or n >= self.ngens():
            raise ValueError, "Generator %s not defined."%n
        return self.__gens[n]

    def _monomial_order_function(self):
        return self.__monomial_order_function

    def __call__(self, x):
        """
        Coerce x into this multivariate polynomial ring.
        """
        if isinstance(x, multi_polynomial.MPolynomial_polylist) and x.parent() == self:
            return x
        y = self.base_ring()(x)
        return multi_polynomial.MPolynomial_polylist(self, [[y, list(self.__zero_list)]])
