# William Stein's comments below are prefixed with #WS:

#WS: I'm going to comment up your code with ideas to improve it
#WS: instead of just rewriting the whole thing.  This way we'll
#WS: both learn a lot quickly.   I did have to rewrite a tiny
#WS: amount below in order to get this to work with SAGE (as
#WS: part of multi_polynomial.


#############################################################
##### uses the "distributive representation" of polynomials
#####   [[c1,[e1,...,er]],...] <-> c1*x1^e1*...*xr^er+...
##### to add and mutiply polynomials
##### we call such a representation a "polylist"
#####6-19-2005,wdj@usna.edu
################################################################

"""
EXAMPLE: (1) put polylist.fcns.py in
/home/wdj/pythonstuff/math/william_stein/sage-v0.3-2005-06-02/local/lib/python2.4/site-packages/sage/rings
         (2) type:
> cd /home/wdj/pythonstuff/math/william_stein/sage-v0.3-2005-06-02
> bin/sage
sage: import sage.rings.polylist_fcns as PList
sage: polylist1=[[2,[2,3]],[3,[1,2]],[4,[2,1]]]
sage: polylist2=[[-1,[1,3]],[-3,[1,2]],[2,[1,1]]]
sage: P1=PList.PolyList(polylist1)
sage: P2=PList.PolyList(polylist2)
sage: P1+P2
      PolyList:[[2, [2, 3]], [4, [2, 1]], [-1, [1, 3]], [0, [1, 2]], [2, [1, 1]]]
sage: P3=P1+P2
sage: print P3
 2x_0^2x_1^3 + 4x_0^2x_1 + -1x_0x_1^3x_0x_1^2 + 2x_0x_1
sage: P4=P1*P2
sage: P4
      PolyList with distributtive representation:[[-2, [3, 6]], [-6, [3, 5]], [0, [3, 4]], [-3, [2, 5]], [-9, [2, 4]], [6, [2, 3]], [-12, [3, 3]], [8, [3, 2]]]
sage: print P4
 -2x_0^3x_1^6 + -6x_0^3x_1^5x_0^3x_1^4 + -3x_0^2x_1^5 + -9x_0^2x_1^4 + 6x_0^2x_1^3 + -12x_0^3x_1^3 + 8x_0^3x_1^2
sage: polylist5=[[2,[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20]],[3,[1,0,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20]],[4,[2,1,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20]]]
sage: P5=PList.PolyList(polylist5)
sage: P5
      PolyList with distributive representation:[[2, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]], [3, [1, 0, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]], [4, [2, 1, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]]]
sage: print P5
 2x_0x_1^2x_2^3x_3^4x_4^5x_5^6x_6^7x_7^8x_8^9x_9^10x_10^11x_11^12x_12^13x_13^14x_14^15x_15^16x_16^17x_17^18x_18^19x_19^20 + 3x_0x_1^0x_2^3x_3^4x_4^5x_5^6x_6^7x_7^8x_8^9x_9^10x_10^11x_11^12x_12^13x_13^14x_14^15x_15^16x_16^17x_17^18x_18^19x_19^20 + 4x_0^2x_1x_2^3x_3^4x_4^5x_5^6x_6^7x_7^8x_8^9x_9^10x_10^11x_11^12x_12^13x_13^14x_14^15x_15^16x_16^17x_17^18x_18^19x_19^20
sage:        

"""

#WS: What is in sage.linalg.vector_space that you need which is not in sage.all?

#WS: You can do "from sage.all import *" (or similar) inside any part of SAGE, since
#WS: it sets up a infinite inclusion loop.
#from sage.linalg.vector_space import *
#from sage.all import *

def hi():
   ## for testing purposes...
   print "Hello. Polynomial list functions loaded."

def flatten(lists):
   # used to convert [[row1],[row2],...,[rowk]] matrix notation
   # to [row1,...,rowk]
   n=len(lists[1])
   k=len(lists)
   ans=lists[0]
   for i in range(k-1):
      ans+=lists[i+1]
   return ans
   #WS: Replace this whole function by the single line
   #WS:  return sum(lists,[])
   #WS: which has the advantage that it works even if the row lengths
   #WS: are different, and is faster.

#WS: Regarding union and intersection, it would probably be much better
#WS: to use the new set type, which was introduced in Python 2.4.  It
#WS: provides extensive "set" functionality, and is a highly optimized C
#WS: implementation.  In particular, you shouldn't use your Set class
#WS: below, but instead the Python 2.3 builtin set class.

def list_union(self, other):
   # Input: self, other - any sequence
   # Output: their union
       res = self[:]
       for x in other:
          if not(x in res):
             res.append(x)
       return res

def list_intersection(self, other):
   # Input: self, other - any sequence
   # Output: their intersection
       res = self[:]
       cap=[]
       for x in other:
          if x in res:
             cap.append(x)
       return cap


def list_filtered(list,boolprop):
   # list is a list
   # boolprop is a Boolean-valued fcn on list
   ans=[]
   for x in list:
      if boolprop(x):
         ans.append(x)
   return ans
   #WS: This list function is already builtin to Python and is called "filter".
   #WS: Also, the current recomendation that I've heard is to use list comprehension instead of filter,
   #WS: i.e., the notation
   #WS:          [x for x in list if boolprop(x)]
   #WS: for what you have above.  This also looks very similar to MAGMA and is easier to understand.

def list_positions(list,x):
   # list is a list
   # x is anything
   # returns list of indices i where x = list[i]
   ans=[]
   n=len(list)
   for i in range(n):
      if x==list[i]:
         ans.append(i)
   return ans
   #WS: If you do this more than a few times on a single list it would be better that this list is sorted.

def list_total(list):
   # list is a list of numbers; returns their sum
   ans=0
   n=len(list)
   for i in range(n):
      ans+=list[i]
   return ans
   #WS: Change this function to
   #WS:     return sum(list,0)
   #WS: The extra 0 is so the empty list returns a sum of 0.

def vector_add(list1,list2):
   # list1,list2 are lists of numbers of same length
   # returns the component-wise sum, as though they were vectors
   n=len(list1)           #does not check if lengths agree
   return [(list1[i]+list2[i]) for i in range(n)]


#################################################################
######## set-theoretic stuff
######## copied from Learning Python, 2nd ed, chapter 23

#WS: As mentioned above, use the builtin set class.

class Set:
   def __init__(self, value = []):    # constructor
       self.data = []                 # manages a list
       self.concat(value)

   def intersect(self, other):        # other is any sequence
       res = []                       # self is the subject
       for x in self.data:
           if x in other:             # pick common items
               res.append(x)
       return Set(res)                # return a new Set

   def union(self, other):            # other is any sequence
       res = self.data[:]             # copy of my list
       for x in other:                # add items in other
           if not x in res:
               res.append(x)
       return Set(res)

   def concat(self, value):           # value: list, Set...
       for x in value:                # removes duplicates
          if not x in self.data:
               self.data.append(x)

   def __len__(self):          return len(self.data)        # len(self)
   def __getitem__(self, key): return self.data[key]        # self[i]
   def __and__(self, other):   return self.intersect(other) # self & other
   def __or__(self, other):    return self.union(other)     # self | other
   def __repr__(self):         return 'Set:' + `self.data`  # print

#################################################################

def lex_monord(left,right):
         ## left,right are monomial polylists
         ## returns comp fcn (i.e., takes values 1, 0, -1) on polylists
         #### EXAMPLES
         # sage: from sage.rings.polylist_fcns import *
         # sage: polylist1=[[2,[2,3]],[3,[1,2]],[4,[2,1]]]
         # sage: polylist1.sort(lex_monord)
         # sage: polylist1
         #  [[2, [2, 3]], [4, [2, 1]], [3, [1, 2]]]
         # sage: polylist1.sort(degrevlex_monord)
         # sage: polylist1
         #  [[2, [2, 3]], [4, [2, 1]], [3, [1, 2]]]
           p=left[1]
           q=right[1]
           if p==q:
              return 0
           ans=p<q;    #WS: that semicolon is not needed
           return int(2*int(ans)-1)

#WS: The comments to your functions are not in the standard style
#WS: used for Python, so I'd have to change them all...

def degrevlex_monord(left,right):
         ## left,right are monomial polylists
         ## returns comp fcn (i.e., takes values 1, 0, -1) on polylists
         #### EXAMPLES
         # sage: from sage.rings.polylist_fcns import *
         # sage: polylist1=[[2,[2,3]],[3,[1,2]],[4,[2,1]]]
         # sage: polylist1.sort(lex_monord)
         # sage: polylist1
         #  [[2, [2, 3]], [4, [2, 1]], [3, [1, 2]]]
         # sage: polylist1.sort(degrevlex_monord)
         # sage: polylist1
         #  [[2, [2, 3]], [4, [2, 1]], [3, [1, 2]]]
           p=left[1]
           q=right[1]
           r=len(p)  # assume this = len(q)
           V=VectorSpace(QQ,r)
           p=V.vector(p)
           q=V.vector(q)
           M=MatrixSpace(QQ,r,r)
           B=[[1 for i in range(r)]]
           for i in range(r-1):
              B.append(standard_basis_vector(-1,r-i,r))
           A=M.matrix(flatten(B))      # this is the matrix repn of the degrevlex
           p=p*A.transpose()           # monomial ordering - see p52 of 
           q=q*A.transpose()           # Computational Commutative Algebra 1
                                       # M Kreuzer, L Robbiano
           if p==q:
              return 0
           ans=p<q;
           return int(2*int(ans)-1)


def coeffs_polylist(polylist):
         n=len(polylist)
         coeffs=[polylist[i][0] for i in range(n)]
         return coeffs

def expons_polylist(polylist):
         n=len(polylist)
         expons=[polylist[i][1] for i in range(n)]
         return expons

def AddPolylists(left, right,monord=lex_monord):
        ##### uses the "distributive representation" of polynomials
        #left=[[c1,[e1,...,er]],...] <-> c1*x1^e1*...*xr^er+...
        ####### examples with collection in add and cancellation in mult
        #from sage.rings.polylist_fcns import *
        # >>> polylist1=[[2,[2,3]],[3,[1,2]],[4,[2,1]]]
        # >>> polylist2=[[-1,[1,3]],[-3,[1,2]],[2,[1,1]]]
        # >>> add_polylists(polylist1,polylist2,lex_monord)
        #     [[2, [2, 3]], [4, [2, 1]], [-1, [1, 3]], [0, [1, 2]], [2, [1, 1]]]
        n1=len(left)
        n2=len(right)
        if n1==1 and left[0][0]==0:
           return right
        if n2==1 and right[0][0]==0:
           return left
        coeffs1=coeffs_polylist(left)
        coeffs2=coeffs_polylist(right)
        expons1=expons_polylist(left)
        expons2=expons_polylist(right)
        newexpons=Set(expons1).union(Set(expons2))
        dblexpons=list_intersection(expons1,expons2)
         # these are the monomials which occur in both polys
        newexponslist=[x for x in newexpons]
         #newexponslist.sort(monord)
         # these are all the pure monomials which occur in either polys
        n=len(newexponslist)
        newpolylist=[]
        for i in range(n):
              if newexponslist[i] in dblexpons:
                 i1=expons1.index(newexponslist[i])
                 i2=expons2.index(newexponslist[i])
                 coeff=coeffs1[i1]+coeffs2[i2]                # new coeff (when pure monomials match)
                 newpolylist.append([coeff,newexponslist[i]]) # leave in if coeff=0???
              if not(newexponslist[i] in dblexpons):
                 if (newexponslist[i] in expons1):
                    i1=expons1.index(newexponslist[i])
                    newpolylist.append([coeffs1[i1],newexponslist[i]])
                 if (newexponslist[i] in expons2):
                    i2=expons2.index(newexponslist[i])
                    newpolylist.append([coeffs2[i2],newexponslist[i]])
        newpolylist.sort(monord)
        return newpolylist

def MultPolylists(left, right,monord=lex_monord):
        ##### uses the "distributive representation" of polynomials
        #left=[[c1,[e1,...,er]],...] <-> c1*x1^e1*...*xr^er+...
        #EXAMPLE:
        ####### examples with collection in add and cancellation in mult
        #from sage.rings.polylist_fcns import *
        # >>> polylist1=[[2,[2,3]],[3,[1,2]],[4,[2,1]]]
        # >>> polylist2=[[-1,[1,3]],[-3,[1,2]],[2,[1,1]]]
        # >>> mult_polylists(polylist1,polylist2)
        # [[-2, [3, 6]], [-6, [3, 5]], [0, [3, 4]], [-3, [2, 5]],
        #  [-9, [2, 4]], [6, [2, 3]], [-12, [3, 3]], [8, [3, 2]]]
        #
        #
        zero=0                        # 0 in **coeff ring** of polys
        n1=len(left)
        n2=len(right)
        r=len(left[0][1])                 # no checking if all expons of left,right have same size
        coeffs1=coeffs_polylist(left)
        coeffs2=coeffs_polylist(right)
        expons1=expons_polylist(left)
        expons2=expons_polylist(right)
        newpolylist=[]
        for lpoly in left:                           # mult together left*right - no collecting
           for rpoly in right:
              newpolylist.append([lpoly[0]*rpoly[0],vector_add(lpoly[1],rpoly[1])])
        newcoeffs=coeffs_polylist(newpolylist)
        newexpons=expons_polylist(newpolylist)
        multpoly=[]
        expons=[]
        for poly in newpolylist:
           num_occur=newexpons.count(poly[1])         # number times a given "pure" monomial occurs
           if num_occur==1:
              multpoly.append(poly)
              expons.append(poly[1])
           if num_occur>1 and not(poly[1] in expons):    # collecting new "like" terms
              pos=list_positions(newexpons,poly[1])     # positions where pure monomial occurs
              tpoly=[[zero,poly[1]]]                     # 0 (**in coeff ring**) times pure monomial
              for i in pos:
                 tpoly=AddPolylists(tpoly,[newpolylist[i]])
              expons.append(poly[1])
              multpoly.append(tpoly[0])
#       multpoly.sort(monord)
        return multpoly


class PolyList:
      def __init__(self,plist,morder=lex_monord):
         self.distrepn = plist           # the distributive representation
         n=len(self.distrepn)
         self.coefficients=[plist[i][0] for i in range(n)]
         self.exps=[plist[i][1] for i in range(n)]
         self.monomialordering=morder

      def intersect(self, other):        # other is any sequence
       res = []                       # self is the subject
       for x in self.distrepn:
           if x in other:             # pick common items
               res.append(x)
       return PolyList(res)                # return a new Set

      def union(self, other):            # other is any sequence
       res = self.distrepn[:]             # copy of my list
       for x in other:                # add items in other
           if not x in res:
               res.append(x)
       return PolyList(res)

      def concat(self, value):           # value: list, Set...
       for x in value:                # removes duplicates
          if not x in self.distrepn:
               self.distrepn.append(x)

      def __len__(self):          return len(self.distrepn)        # len(self)

      def __getitem__(self, key): return self.distrepn[key]        # self[i]

      def __or__(self, other):    return self.union(other)     # self | other

      def __repr__(self):         return 'PolyList with distributive representation:' + `self.distrepn`  # print

      def __str__(self):
         nterms=len(self)
         nvars=len(self.exps[0])
         poly=""
         for i in range(nterms):
            if (i>0 and self.distrepn[i][0]==0):
               pass
            elif (i>0 and self.distrepn[i][0]<>0):
               poly=poly+" + %g"%self.distrepn[i][0]
            else:
               poly=poly+" %g"%self.distrepn[i][0]
            for j in range(nvars):
               if self.distrepn[i][1][j]==1:
                  poly=poly+"x_%d" %j
               else:
                  poly=poly+"x_%d^%s" %(j,self.distrepn[i][1][j])
         return poly

      def dirac_delta(i,j):
         ## i,j, objects
         ## returns 1 if i=j, 0 otherwise
         return int(i==j)

      def standard_basis_vector(a,i,n):
         # returns basis n-vector with a in ith position
         # a an integer
         return [a*int(j==i-1) for j in range(n)]

      def add_polylists(left, right,monord =lex_monord):
        ##### uses the "distributive representation" of polynomials
        #left=[[c1,[e1,...,er]],...] <-> c1*x1^e1*...*xr^er+...
        ####### examples with collection in add and cancellation in mult
        #from sage.rings.polylist_fcns import *
        # >>> polylist1=[[2,[2,3]],[3,[1,2]],[4,[2,1]]]
        # >>> polylist2=[[-1,[1,3]],[-3,[1,2]],[2,[1,1]]]
        # >>> add_polylists(polylist1,polylist2,lex_monord)
        #     [[2, [2, 3]], [4, [2, 1]], [-1, [1, 3]], [0, [1, 2]], [2, [1, 1]]]
        n1=len(left)
        n2=len(right)
        if n1==1 and left.distrepn[0][0]==0:
           return right
        if n2==1 and right.distrepn[0][0]==0:
           return left
        coeffs1=left.coefficients
        coeffs2=right.coefficients
        expons1=left.exps
        expons2=right.exps
        newexpons=Set(expons1).union(Set(expons2))
        dblexpons=list_intersection(expons1,expons2)
         # these are the monomials which occur in both polys
        newexponslist=[x for x in newexpons]
         # these are all the pure monomials which occur in either polys, sorted
        n=len(newexponslist)
        newpolylist=[]
        for i in range(n):
              if newexponslist[i] in dblexpons:
                 i1=expons1.index(newexponslist[i])
                 i2=expons2.index(newexponslist[i])
                 coeff=coeffs1[i1]+coeffs2[i2]                # new coeff (when pure monomials match)
                 newpolylist.append([coeff,newexponslist[i]]) # leave in if coeff=0???
              if not(newexponslist[i] in dblexpons):
                 if (newexponslist[i] in expons1):
                    i1=expons1.index(newexponslist[i])
                    newpolylist.append([coeffs1[i1],newexponslist[i]])
                 if (newexponslist[i] in expons2):
                    i2=expons2.index(newexponslist[i])
                    newpolylist.append([coeffs2[i2],newexponslist[i]])
        newpolylist.sort(monord)
        return PolyList(newpolylist)
     
      def __add__(self, other):   return self.add_polylists(other) # self + other

      def mult_polylists( left, right,monord =lex_monord ):
        ##### uses the "distributive representation" of polynomials
        #left=[[c1,[e1,...,er]],...] <-> c1*x1^e1*...*xr^er+...
        #EXAMPLE:
        ####### examples with collection in add and cancellation in mult
        #from sage.rings.polylist_fcns import *
        # >>> polylist1=[[2,[2,3]],[3,[1,2]],[4,[2,1]]]
        # >>> polylist2=[[-1,[1,3]],[-3,[1,2]],[2,[1,1]]]
        # >>> mult_polylists(polylist1,polylist2)
        # [[-2, [3, 6]], [-6, [3, 5]], [[0, [3, 4]]], [-3, [2, 5]],
        #  [-9, [2, 4]], [6, [2, 3]], [-12, [3, 3]], [8, [3, 2]]]
        #
        #
        zero=0                        # 0 in **coeff ring** of polys
        n1=len(left)
        n2=len(right)
        r=len(left.distrepn[0][1])        # no checking if all expons of left,right have same size
        coeffs1=left.coefficients
        coeffs2=right.coefficients
        expons1=left.exps
        expons2=right.exps
        newpolylist=[]
        for lpoly in left.distrepn:                     # mult together left*right - no collecting
           for rpoly in right.distrepn:
              newpolylist.append([lpoly[0]*rpoly[0],vector_add(lpoly[1],rpoly[1])])
        newcoeffs=PolyList(newpolylist).coefficients
        newexpons=PolyList(newpolylist).exps
        multpoly=[]
        expons=[]
        for poly in newpolylist:
           num_occur=newexpons.count(poly[1])         # number times a given "pure" monomial occurs
           if num_occur==1:
              multpoly.append(poly)
              expons.append(poly[1])
           if num_occur>1 and not(poly[1] in expons):    # collecting new "like" terms
              pos=list_positions(newexpons,poly[1])     # positions where pure monomial occurs
              tpoly=[[zero,poly[1]]]                     # 0 (**in coeff ring**) times pure monomial
              for i in pos:
                 tpoly=PolyList.add_polylists(PolyList(tpoly),PolyList([newpolylist[i]]))
              expons.append(poly[1])
              multpoly.append(tpoly)
#       list(multpoly).sort(monord)
        return PolyList(multpoly)

      def __mul__(self, other):   return PolyList(MultPolylists(self.distrepn,other.distrepn)) # self * other






           
